using System.Collections.Generic;
using UnityEngine;

namespace SwimmingInVR.Scripts
{
    /// <summary>
    ///     Enables swimming in VR.
    ///     Swimming is possible by moving the controllers in the opposite direction of the target direction
    ///     while pressing the swimming button on both controllers.
    ///     Gravity and water resistance affect the player.
    /// </summary>
    public class SwimmingPlayerController : MonoBehaviour
    {
        /// <summary>
        ///     Gravity in the virtual environment (default is the gravity on earth).
        /// </summary>
        private const float FGravityForce = 9.81f;

        /// <summary>
        ///     The lower the value, the more frequently swimming movements must be performed.
        /// </summary>
        [Header("Swimming")] [SerializeField] private float fSwimmingForce;

        [SerializeField] private float fWaterResistance;

        /// <summary>
        ///     If the Controller movements is smaller than this value, the movement is not counted.
        /// </summary>
        [SerializeField] private float fHandMovementDeadZone;

        [SerializeField] private float fMinTimeBetweenSwimmingMovements;
        [SerializeField] private Transform trackingSpaceTransform;
        [SerializeField] private OVRInput.Button swimmingButton = OVRInput.Button.PrimaryHandTrigger;

        /// <summary>
        ///     How much is gravity reduced under water. Is multiplied with the gravity.
        /// </summary>
        [Header("Gravity")] [SerializeField] private float fUnderwaterGravityScale;

        /// <summary>
        ///     Each time you swim, one of these sounds is randomly selected and played.
        /// </summary>
        [Header("Sounds")] [SerializeField] private List<AudioClip> swimmingSounds;

        [SerializeField] private AudioSource audioSource;
        private float _currentWaitTimeBetweenSwimmingMovements;
        private Rigidbody _rigidbody;
        private Vector3 _v3CurrentSwimmingDirection;

        private void Awake()
        {
            // Rigidbody setup

            _rigidbody = GetComponent<Rigidbody>();
            _rigidbody.useGravity = false;
            _v3CurrentSwimmingDirection = Vector3.zero;
            _rigidbody.velocity = Vector3.zero;
            _rigidbody.angularVelocity = Vector3.zero;
            _rigidbody.isKinematic = false;

            AddWaterResistance();
            AddGravity();
        }

        private void FixedUpdate()
        {
            _currentWaitTimeBetweenSwimmingMovements += Time.deltaTime;

            if (OVRInput.Get(swimmingButton, OVRInput.Controller.RTouch) &&
                OVRInput.Get(swimmingButton, OVRInput.Controller.LTouch))
            {
                var leftControllerDirection = OVRInput.GetLocalControllerVelocity(OVRInput.Controller.LTouch);
                var rightControllerDirection = OVRInput.GetLocalControllerVelocity(OVRInput.Controller.RTouch);
                var leftAndRightControllerVelocitySum = leftControllerDirection + rightControllerDirection;

                var oppositeVelocitySum = -leftAndRightControllerVelocitySum;
                if (oppositeVelocitySum.sqrMagnitude > fHandMovementDeadZone * fHandMovementDeadZone &&
                    _currentWaitTimeBetweenSwimmingMovements > fMinTimeBetweenSwimmingMovements)
                {
                    var velocityInWorldSpace = trackingSpaceTransform.TransformDirection(oppositeVelocitySum);
                    Swim(velocityInWorldSpace);
                }
            }

            AddWaterResistance();
            AddGravity();
        }

        private void Swim(Vector3 swimmingVelocity)
        {
            _rigidbody.AddForce(swimmingVelocity * fSwimmingForce, ForceMode.Impulse);
            _v3CurrentSwimmingDirection = swimmingVelocity.normalized;
            _currentWaitTimeBetweenSwimmingMovements = 0;

            if (!audioSource) return;
            audioSource.clip = swimmingSounds[Random.Range(0, swimmingSounds.Count)];
            audioSource.Play();
        }

        private void AddWaterResistance()
        {
            if (_rigidbody.velocity.sqrMagnitude > 0.01f && _v3CurrentSwimmingDirection != Vector3.zero)
                _rigidbody.AddForce(-_rigidbody.velocity * fWaterResistance, ForceMode.Acceleration);

            else
                _v3CurrentSwimmingDirection = Vector3.zero;
        }

        private void AddGravity()
        {
            var gravity = FGravityForce * fUnderwaterGravityScale * Vector3.up;
            _rigidbody.AddForce(-gravity);
        }
    }
}